@@ -37,10 +37,6 @@
 
 #include "svn_private_config.h"
 
-/* A handy constant */
-static const svn_opt_revision_t unspecified_revision
-  = { svn_opt_revision_unspecified, { 0 } };
-
 
 /*** Code. ***/
 
@@ -101,15 +97,6 @@ run_merge(svn_boolean_t two_sources_specified,
           && (first_range_end.kind == svn_opt_revision_unspecified))
         {
           ranges_to_merge = NULL;
-
-          /* This must be a 'sync' merge so check branch relationship. */
-          if (opt_state->verbose)
-            SVN_ERR(svn_cmdline_printf(
-                      scratch_pool, _("--- Checking branch relationship\n")));
-          SVN_ERR_W(svn_cl__check_related_source_and_target(
-                      sourcepath1, &peg_revision1,
-                      targetpath, &unspecified_revision, ctx, scratch_pool),
-                _("Source and target must be different but related branches"));
         }
 
       if (opt_state->verbose)
@@ -136,6 +123,11 @@ run_merge(svn_boolean_t two_sources_specified,
                                 _("Merge sources must both be "
                                   "either paths or URLs"));
 
+      if (svn_path_is_url(targetpath))
+        return svn_error_createf(SVN_ERR_CL_ARG_PARSING_ERROR, NULL,
+                                 _("Merge target '%s' must be a local path "
+                                   "but looks like a URL"), targetpath);
+
       if (opt_state->verbose)
         SVN_ERR(svn_cmdline_printf(scratch_pool, _("--- Merging\n")));
       merge_err = svn_client_merge5(sourcepath1,
